using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;

namespace Stopper {
	public class EyelidBehaviorEditor : MVRScript {

        string savePath = "Saves\\EyelidBehavior";
		string fileExt = "eyelidbehavior";

        private DAZMeshEyelidControl eyelidControl;
        private List<string> morphNames = new List<string>();
        private HashSet<DAZMorph> modifiedMorphs = new HashSet<DAZMorph>();

        private bool morphsMeedUpdate = false;
        private bool needRebuildMorphs = false;
		public override void Init() {
			eyelidControl = containingAtom.GetComponentInChildren<DAZMeshEyelidControl>();
            RefreshMorphList();

            // Preset code was adapted from AcidBubbles' Collider Editor
			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});

            // Blink Controls
            AddSlider("Blink Space Min", 1f, 0.001f, 10, (val) => {
                eyelidControl.blinkSpaceMin = val;
            }, true, "F2", false);

            AddSlider("Blink Space Max", 7f, 0.001f, 10, (val) => {
                eyelidControl.blinkSpaceMax = val;
            }, true, "F2", false);

            AddSlider("Blink Time Min", 0.1f, 0.001f, 1f, (val) => {
                eyelidControl.blinkTimeMin = val;
            }, true, "F3", false);

            AddSlider("Blink Time Max", 0.4f, 0.001f, 1f, (val) => {
                eyelidControl.blinkTimeMin = val;
            }, true, "F3", false);

            AddSlider("Blink Up/Down Ratio", 0.4f, 0.00f, 1f, (val) => {
                eyelidControl.blinkDownUpRatio = val;
            }, true, "F3", false);

            AddSlider("Blink Bottom Eyelid Factor", 0.5f, 0.00f, 1f, (val) => {
                eyelidControl.blinkBottomEyelidFactor = val;
            }, true, "F3", false);

            // Look controls
            AddSlider("Look Up Top Eyelid Factor", 3f, 0f, 5f, (val) => {
                eyelidControl.lookUpTopEyelidFactor = val;
            }, true, "F2", false);

            AddSlider("Look Up Bottom Eyelid Factor", 1f, 0f, 5f, (val) => {
                eyelidControl.lookUpBottomEyelidFactor = val;
            }, true, "F2", false);

            AddSlider("Look Down Top Eyelid Factor", 1.5f, 0f, 5f, (val) => {
                eyelidControl.lookDownTopEyelidFactor = val;
            }, true, "F2", false);

            AddSlider("Look Down Bottom Eyelid Factor", 4f, 0f, 5f, (val) => {
                eyelidControl.lookDownBottomEyelidFactor = val;
            }, true, "F2", false);

            SetUpMorphChooser("LeftTopEyelidDownMorphName", "Left Top Eyelid Down Morph", eyelidControl.LeftTopEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.LeftTopEyelidDownMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("RightTopEyelidDownMorphName", "Right Top Eyelid Down Morph", eyelidControl.RightTopEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.RightTopEyelidDownMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("LeftBottomEyelidDownMorphName", "Left Bottom Eyelid Down Morph", eyelidControl.LeftBottomEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.LeftBottomEyelidDownMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("RightBottomEyelidDownMorphName", "Right Bottom Eyelid Down Morph", eyelidControl.RightBottomEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.RightBottomEyelidDownMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("LeftTopEyelidUpMorphName", "Left Top Eyelid Up Morph", eyelidControl.LeftTopEyelidUpMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.LeftTopEyelidUpMorphName = morphName;
                    }
                }
            );

             SetUpMorphChooser("RightTopEyelidUpMorphName", "Right Top Eyelid Up Morph", eyelidControl.RightTopEyelidUpMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.RightTopEyelidUpMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("LeftBottomEyelidDownMorphName", "Left Bottom Eyelid Down Morph", eyelidControl.LeftBottomEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.LeftBottomEyelidDownMorphName = morphName;
                    }
                }
            );

            SetUpMorphChooser("RightBottomEyelidDownMorphName", "Right Bottom Eyelid Down Morph", eyelidControl.RightBottomEyelidDownMorphName,
                (string morphName) => {
                    if(SetUpMorph(morphName)) {
                        eyelidControl.RightBottomEyelidDownMorphName = morphName;
                    }
                }
            );

		}

        private void Update() {
            if(needRebuildMorphs) {
                RebuildMorphs();
                needRebuildMorphs = false;        
            }

            if(morphsMeedUpdate) {
                
                UpdateMorphs();
                morphsMeedUpdate = false;
            }
        }

        private void SetUpMorphChooser(string storeName, string displayName, string defaultVal, JSONStorableStringChooser.SetStringCallback callback) {
            var chooser = new JSONStorableStringChooser(storeName, morphNames, defaultVal, displayName, callback);
		
			RegisterStringChooser(chooser);
			var popup = CreateFilterablePopup(chooser, true);
			popup.popup.onOpenPopupHandlers += () => {
                RefreshMorphList();
				chooser.choices = morphNames;
			};
        }

        private bool SetUpMorph(string morphName) {
            var morph = eyelidControl.morphBank.GetMorph(morphName);
            if(morph == null) {
                SuperController.LogError($"Could not find morph {morphName}");
                return false;
            }

            if(morph.isRuntime) {
                morph.isRuntime = false;
                if(!morph.deltasLoaded) {
                    morph.LoadDeltas();
                }
                modifiedMorphs.Add(morph);
                needRebuildMorphs = true;
            }
            
            morphsMeedUpdate = true;
            return true;
        }

        private void RebuildMorphs() {
            eyelidControl.morphBank.RebuildAllLookups();
        }
        private void UpdateMorphs() {
            var morphBank = eyelidControl.morphBank;
            eyelidControl.morphBank = null;
            eyelidControl.morphBank = morphBank;
        }

        private JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, string valueFormat = "F2", bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.valueFormat = valueFormat;
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return storableFloat;
        }

        private void RefreshMorphList(){
            morphNames = eyelidControl.morphBank.morphs.Select((m) => m.morphName).ToList();
        }
		
		
		private void OnDisable()
		{
            foreach(var entry in floatParams) {
                var storableFloat = entry.Value;
                storableFloat.setCallbackFunction(storableFloat.defaultVal);
            }

            foreach(var entry in stringChooserParams) {
                var stringChooser = entry.Value;
                stringChooser.setCallbackFunction(stringChooser.defaultVal);
            }
            UpdateMorphs();
		}

        private void OnEnable()
		{
            foreach(var entry in floatParams) {
                var storableFloat = entry.Value;
                storableFloat.setCallbackFunction(storableFloat.val);
            }

            foreach(var entry in stringChooserParams) {
                var stringChooser = entry.Value;
                stringChooser.setCallbackFunction(stringChooser.val);
            }
		}

        private void OnDestroy(){
            foreach(var morph in modifiedMorphs) {
                morph.isRuntime = true;
            }
            RebuildMorphs();
        }

        private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);
			
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}
	}
}